/*
 * Copyright 04/24/96 Sun Microsystems, Inc. All Rights Reserved
 *
 * This example program shows how to use the DLPI interface
 * to create and establish a LAPB link using SunLink X.25
 *
 * Its arguments are :
 *      the link_id to be used
 *      the name of the serial driver to use (zsh, hsi, wloop ...)
 *      the type of link (0 for DTE, 1 for DCE)
 *
 * ex : to establish a link between wloop loopback driver :
 *   lapb 38 /dev/wloop0 0 to create DTE side with link_id 38 over wloop0 driver
 *   lapb 39 /dev/wloop1 1 to create DCE side with link_id 39 over wloop1 driver
 *
 */

#include <errno.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/dlpi.h>
#include <netdlc/uint.h>
#include <netdlc/ll_control.h>
#include <netdlc/ll_proto.h>

/* Prototypes of functions used for the various stages of DLPI */
int build_stream_arch(char * device, int link_id, int DCE_flag);
void send_dl_attach(int fd, int link_id);
void send_dl_bind(int fd);
void wait_and_ack_dl_connect_ind(int fd);
void send_data(int fd);
void recv_data(int fd);
void send_dl_disconnect_req(int fd);
void wait_and_ack_dl_disconnect_ind(int fd);
void wait_dlpi_prim(int fd, int exp_prim, char *ack_prim_str);
void usage(void);

/* these structures are used to send and received the DLPI messages */
struct strioctl strio;
struct strbuf ctlbuf;

main(int argc, char *argv[])
{

	int lapb_fd;
	int link_id, DCE_flag;
	char* device;

	/* parse the arguments */
	if (argc != 4)
		usage();

	link_id = atoi(argv[1]);
	device = argv[2];
	DCE_flag = atoi(argv[3]);

	/* build the STREAMS architecture used for LAPB protocol
	 * Rq : This operation is not part of the DLPI standard
	 */
	lapb_fd = build_stream_arch(device, link_id, DCE_flag);

	/* Now send DL_ATTACH_REQ to specify we want to use ppa "link_id" */
	send_dl_attach(lapb_fd, link_id);

	/* Now this stream must be binded
	 * The bind operation allows to specify to the LAPB protocol
	 * driver the parameters that we are going to use.
	 * after this operation the stream we'll be ready for connection
	 * establishment phase.
	 */
	send_dl_bind(lapb_fd);

	/* Now the connection must be established 
	 * The LAPB protocol automatically do this as soon
	 * as both side are binded.
	 * On success it will send DL_CONNECT_IND messages 
	 */
	wait_and_ack_dl_connect_ind(lapb_fd);

	/* No the link is UP and ready for data transmission.
	 * Here we perform Full-duplex data transmission :
	 * each side of the link send one data frame and wait for the
	 * frame sent by the remote 
	 * The M_DATA message is used during data transfer
	 */
	send_data(lapb_fd);
	recv_data(lapb_fd);

	/* Now we can disconnect the stream.
	 * during disconenction phase one side has to initiate the
	 * disconnection, the other side will wait for a 
	 * disconnection indication.
	 * In this example we have choosen the DCE to initiate 
	 * the disconnection (but the reverse way is possible) 
	 */
	if (DCE_flag)
		send_dl_disconnect_req(lapb_fd);
	else
		wait_and_ack_dl_disconnect_ind(lapb_fd);

	/* Now the stream is disconnected and in state BOUNDED.
	 * we can reuse it to establish and other connection,
	 * or UNBIND it or just exit.
	 * Exiting will cause the stream to be automatically unbinded
	 * and dismantled (it would also have caused disconnection
	 * if we had not done it previously). 
	 *
	 * in this example we just exit.
	 */
	exit(0);
}

/* This portion of code shows how to build the STREAMS architecture 
 * so as to use the LAPB DLPI interface on top of the 
 * serial device.
 * 
 * In order to be able to use LAPB DLPI interface the stream
 * architecture must be :
 *
 *           |
 *  ------------------
 *  | LAPB driver     |
 *  ------------------
 *           |
 *  ------------------
 *  | wanmod module  |
 *  ------------------
 *           |
 *  ------------------
 *  | serial driver  |
 *  ------------------
 *
 * The serial driver manages the chip that controls the serial line.
 * The wanmod that is "pushed" between the serial driver and the
 * LAPB driver controls the signals sent by the serial driver and
 * informs the LAPB driver when the connection is established at
 * the physical level (i.e. when you plug the cable)
 * The lapb driver implements the LAPB protocol and the DLPI interface.
 *
 * This function uses the link id and device name as input and
 * return the file descriptor opened to lapb device
 */
int build_stream_arch(char * device, int link_id, int DCE_flag)
{
	int dev_fd, muxid;
	int lapb_fd;
	struct ll_snioc snioc;

	/* Open the serial device */
	dev_fd = open(device, O_RDWR);
	if (dev_fd < 0) {
                printf("Cannot open %s : %s\n", device, strerror(errno));
                exit(1);
        }

	/* Open the lapb device in order to access to the LAPB DLPI interface */
	lapb_fd = open("/dev/lapb", O_RDWR);
	if (lapb_fd < 0) {
                perror("Cannot open /dev/lapb");
                exit(1);
        }

	/* PUSH the wanmod module on top of device */
	if (ioctl(dev_fd, I_PUSH, "wanmod") == -1) {
		perror("        **ioctl(I_PUSH wanmod)");
		exit(1);
	}

	/* I_LINK device file (and wanmod) underneath /dev/lapb. */
	if ((muxid = ioctl(lapb_fd, I_LINK, dev_fd)) < 0) {
		perror("	**ioctl(I_LINK)");
		exit (1);
	}
	/* The dev_fd is now linked under LAPB driver, since it is not
	 * usefull anymore for this application we can close it
	 */
	close(dev_fd);

	/*
	 * The STREAMS architecture is now constructed and can be used.
	 * The first step is to send the L_SETPPA ioctls in order to
	 * Set the PPA/snid of the device stream under LAPB to 'link_id'.
	 * This PPA will then be used in the DL_ATTACH ioctl to identify
	 * the link.
	 *
	 * CAUTION : the link_id must be different from the ones used 
	 * in the X.25 configuration. Otherwise the LAPB driver would
	 * consider this link is already defined and return and error EBUSY
	 *
	 * Both L_SETPPA and DL_ATTACH ioctls are described
	 * in the X.25 Programming guide
	 */
	snioc.lli_type = LI_SPPA;
	snioc.lli_ppa = link_id;
	snioc.lli_index = muxid;
	/* this is the time where we decide if we are DCE or DTE */
	snioc.lli_class = (DCE_flag ? LC_LAPBDCE : LC_LAPBDTE);

	strio.ic_cmd = L_SETPPA;
	strio.ic_timout = -1;			/* Infinite timeout */
	strio.ic_len = sizeof(snioc);
	strio.ic_dp = (char *) &snioc;
	if (ioctl(lapb_fd, I_STR, (char *) &strio) < 0) {
		perror("	**ioctl(L_SETPPA)");
		exit (1);
	}

	/* At this moment the L_SETTUNE ioctl can be used 
 	 * To configure LAPB parameters (T1, N2, inactivity timer, ect...)
	 * otherwise default values are used for this LAPB LINK
	 * See X.25 Programming Guide for more information about this ioctl
	 */
	return lapb_fd;
}
	
/* This function attaches a previously build STREAMS architecture
 * using the ppa : link_id 
 */
void send_dl_attach(int fd, int link_id)
{
	dl_attach_req_t dlatt;
	int flags;

	dlatt.dl_primitive = DL_ATTACH_REQ;
	dlatt.dl_ppa = link_id;
	ctlbuf.buf = (char *)&dlatt;
        ctlbuf.len = sizeof(dlatt);
        if (putmsg(fd, &ctlbuf, (struct strbuf *)0, 0) < 0)
        {
               perror("DL_ATTACH put message failed");
               exit(1);
        }

	/* Wait for the OK_ACK message that should acknowledge ATTACH request
	 *
	 * The generic union structure DL_primitives is used to get
	 * messages from the LAPB driver because several types of messages
	 * can be returned depending on the success or failure of the 
	 * operation 
	 */
	wait_dlpi_prim(fd, DL_OK_ACK, "DL_ATTACH_REQ");
}

/* This function BIND a lapb stream 
 * The stream must have been previously attached
 */
void send_dl_bind(int fd)
{
	dl_bind_req_t dlbind;

	dlbind.dl_primitive = DL_BIND_REQ;
	/* No SAP is used by the LAPB protocol, the dl_sap field must
	 * therefore be set with 0 when using lapb driver
	 */
	dlbind.dl_sap = 0;
	/* Maximum number of connection usable on that stream.
	 * Always 1 with LAPB.
	 */
	dlbind.dl_max_conind = 1;
	/* service mode is always Connection Oriented when using LAPB */
	dlbind.dl_service_mode = DL_CODLS;
	dlbind.dl_conn_mgmt = 1;
	dlbind.dl_xidtest_flg = DL_AUTO_XID|DL_AUTO_TEST;
	ctlbuf.buf = (char *)&dlbind;
        ctlbuf.len = sizeof(dlbind);
        if (putmsg(fd, &ctlbuf, (struct strbuf *)0, 0) < 0)
        {
               perror("DL_BIND put message failed");
               exit (1);
        }

	/* Now wait for the DL_BIND_ACK primitive that should acknowledge 
	 * the DL_BIND_REQ
	 */
	wait_dlpi_prim(fd, DL_BIND_ACK, "DL_BIND_REQ");
}

/* This function waits for a DL_CONNECT_IND (connection indication)
 * from the LAPB layer, and send back a DL_CONNECT_RES to accept the
 * connection.
 */
void wait_and_ack_dl_connect_ind(int fd)
{
	union DL_primitives dlprim;
	dl_connect_res_t dlconnres;
	int flags;

	/* Now wait for the DL_CONNECT_IND primitive that should indicate 
	 * that the LAPB has completed connection phase
	 */
        ctlbuf.buf = (char *)&dlprim;
        ctlbuf.maxlen = sizeof(dlprim);
 
        flags = 0;
 
        if (getmsg(fd, &ctlbuf, (struct strbuf *)0, &flags) != 0)
        {
                perror("DL_CONNECT_IND get message failed");
                exit(1);
        }
 
        if (dlprim.dl_primitive == DL_CONNECT_IND)
                printf("DL_CONNECT_IND received : OK\n");
	else
	{
		printf("DL_CONNECT failed\n");
		exit (1);
	}
	/* respond to CONNECT indication by a DL_CONNECT_RES message */
	dlconnres.dl_primitive = DL_CONNECT_RES;
	dlconnres.dl_correlation = dlprim.connect_ind.dl_correlation;
	dlconnres.dl_resp_token = 0;
	ctlbuf.buf = (char *)&dlconnres;
        ctlbuf.len = sizeof(dlconnres);
        if (putmsg(fd, &ctlbuf, (struct strbuf *)0, 0) < 0)
        {
               perror("DL_CONNECT_RES put message failed");
               exit (1);
        }

	/* Wait for the OK_ACK message that should acknowledge
	 * the CONNECT_RES message
	 */
	wait_dlpi_prim(fd, DL_OK_ACK, "DL_CONNECT_RES");
}
 
/* This function send a 100 octet long buffer to the remote
 * by sending a simple M_DATA message to LAPB driver.
 * No ack message in used in this phase.
 */
void send_data(int fd)
{
	struct strbuf databuf;
	char buf[100];

	databuf.len = 100;
	databuf.buf = buf;

	/* send a M_DATA Message */
	if (putmsg(fd, (struct strbuf *)0, &databuf, 0) < 0)
	{
		perror("M_DATA put message failed");
		exit(1);
	}

	/* release allocated buffer */
	printf("M_DATA (100 bytes) sent : OK\n");
}

/* This function wait for data 
 * data are received within M_DATA messages.
 */
void recv_data(int fd)
{
	struct strbuf databuf;
	int flags = 0;
	/* It is the responsability of the application
	 * to decide and know the maximum length of data that can
	 * be recieved. In practical, this length can not be greater than
	 * max length of LAPB frames.
	 * in this example we know that the other side will never 
	 * send frames bigger than 100
	 */
#define MAXDATALEN 100
	char buf[MAXDATALEN];

	databuf.buf = buf;
	databuf.maxlen = MAXDATALEN;
	databuf.len = 0;

	/* wait for a M_DATA Message */
	if (getmsg(fd, 0, &databuf, &flags) < 0)
	{
		perror("M_DATA receiving failed");
		exit(1);
	}

	/* In this example no checking is done on the data received */

	printf("M_DATA (%d bytes) received : OK\n", databuf.len);
}

/* This function send a DISCONNECT_REQ message and wait for its ACK
 */
void send_dl_disconnect_req(int fd)
{
	dl_disconnect_req_t disconnect_req;

	disconnect_req.dl_primitive = DL_DISCONNECT_REQ;
	disconnect_req.dl_reason = 0;
	disconnect_req.dl_correlation = 0;

	ctlbuf.buf = (char *)&disconnect_req;
	ctlbuf.len = sizeof(disconnect_req);
	if (putmsg(fd, &ctlbuf, (struct strbuf *)0, 0) < 0)
	{
		perror("DL_DISCONNECT_REQ put message failed");
		exit(1);
	}
 
	/* wait for OK_ACK */
	wait_dlpi_prim(fd, DL_OK_ACK, "DL_DISCONNECT_REQ");
}

/* This function waits for a DISCONNECT_IND
 */
void wait_and_ack_dl_disconnect_ind(fd)
{
	wait_dlpi_prim(fd, DL_DISCONNECT_IND, "DL_DISCONNECT_IND");
}

/* This function will wait for a DLPI message acknowledging (exp_prim)
 * a previously sent DLPI primitive (ack_prim_str)
 */
void wait_dlpi_prim(int fd, int exp_prim, char *ack_prim_str)
{
	union DL_primitives dlprim;
	struct strbuf ctlbuf;
	int flags=0;

	/* wait for exp_prim */
	ctlbuf.buf = (char *)&dlprim;
	ctlbuf.maxlen = sizeof(dlprim);
	if (getmsg(fd, &ctlbuf, (struct strbuf *)0, &flags) != 0)
	{
		printf("%s : ", ack_prim_str);
		perror(" get message failed");
		exit(1);
	}
 
	if(dlprim.dl_primitive == exp_prim)
	{
		/* success */
		printf("%s : OK \n",ack_prim_str);
		return;
	}

	switch(dlprim.dl_primitive)
	{
	    case DL_ERROR_ACK:
		printf("Error after %s : DL_ERROR_ACK was received\n",
					ack_prim_str);
		printf("     dl_errno is : %d\n", dlprim.error_ack.dl_errno);
		exit(1);
	    default:
		printf("Error after %s : Unexpected primitive received: %d\n",
					 dlprim.dl_primitive);
		exit(1);
	}
}

/* This function prints USAGE and exit
 *
 */
void usage(void)
{
    printf("USAGE : lapb <link_id> <device> 0 : to use lapb as a DTE\n");
    printf("        lapb <link_id> <device> 1 : to use lapb as a DCE\n\n");
    printf("Ex : to build a LAPB link between loopbackdriver use :\n");
    printf("     lapb 38 /dev/wloop0 0\n");
    printf(" and lapb 39 /dev/wloop1 1\n");
    exit(1);
}
