/*
 * Copyright 04/24/96 Sun Microsystems, Inc. All Rights Reserved
 *
 * This example program shows how to use the DLPI interface
 * to create and establish LLC2 connections using X.25.
 *
 * Its arguments are :
 *      the link_id to be used
 *      the name of the LAN driver to use (/dev/le, ...)
 *	A flag to specify if this is the calling side or the responder side;
 *	  a value of zero means responder, anything else means initiator.
 *
 * ex : to establish an llc2 connection between two stations using ethernet:
 *    llc2 28 /dev/le 0 to create the responder side with a link id of 28 over
 *	               the built-in ethernet port on the first station
 *    llc2 29 /dev/le 1 to create the initiator side with a link id of 29 over
 *	               the built-in ethernet port on the other station
 *
 * For simplicity, this example program always uses the SAP 0x34  
 * and the first LAN port of any type. This can easily
 * be changed by modifying the initialisation of LSAP and call to attach function.
 *
 * When used as a responder, this example accepts an incoming connection on the 
 * same stream that was used to wait for it. It is then
 * not possible to wait for another connection.
 * It is possible, however, to build a program that is able to use several
 * connections.
 * Such a program would have to make the bind with dl_max_conind set to a value
 * bigger than 1. Then for each incomimg connection, it would have to open a 
 * new stream, attach it, bind it, and do the connect response on that stream.
 * In this case, the initial building of LLC2 stream architecture over the
 * LAN driver only needs to be done once, before listening.
 */

#include <fcntl.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/dlpi.h>
#include <netdlc/uint.h>
#include <netdlc/timer.h>
#include <netdlc/ll_control.h>
#include <netdlc/ll_proto.h>
#include <netdlc/llc2.h>

/* Prototypes of functions used for the various stages of DLPI. */
int build_stream_arch(char * device, int link_id);
void usage(void);
void send_dl_attach(int fd, int link_id);
void send_dl_bind(int fd, int initiator_flag);
void send_dl_connect_req(int fd);
void wait_and_ack_dl_connect_ind(int fd);
void send_data(int fd);
void recv_data(int fd);
void wait_dlpi_prim(int fd, int exp_prim, char *ack_prim_str);

/* These structures are used to send and receive DLPI messages. */
struct strioctl strio;
struct strbuf ctlbuf;

/* The main body of this example.
 */
main(int argc, char *argv[])
{
	int llc2_fd;
	int link_id;
	char *device;
	int initiator_flag;

	/* parse the arguments */
	if (argc != 4)
		usage();

	link_id = atoi(argv[1]);
	device = argv[2];
	initiator_flag = atoi(argv[3]);

	/* Build the stream architecture used for LAPB protocol.
	 * Note: This operation is not part of the DLPI standard.
	 */
	llc2_fd = build_stream_arch(device, link_id);

	/* Now send DL_ATTACH_REQ to specify we want to work on ppa "link_id" */
	send_dl_attach(llc2_fd, link_id);

	/* Now this stream must be bound.
	 * The BIND operation is used to specify the protocol  
	 * to be used, and to initiate or receive connections.
	 */
	send_dl_bind(llc2_fd, initiator_flag);
	
	if (initiator_flag)
	{
		/* This portion of code is run by the CALLING SIDE ONLY. */

		/* Now try to establish connection:
		 * Fill a structure with the dl_connect_req_t and llc_dladdr 
		 * structures concatenated and send it in a single message.
		 * The table conn_struct was previously declared to do this.
		 */
		send_dl_connect_req(llc2_fd);
	}
	else
	{
		/* This portion of code is run by the CALLED SIDE ONLY.
		 * The program waits for the DL_CONNECT_IND message and then sends back
		 * a DL_CONNECT_RES to accept the incoming connection.
		 * (Note: a DL_DISCONNECT_IND could also be used to refuse
		 * this connection)
		 */
		wait_and_ack_dl_connect_ind(llc2_fd);
	}

	/* Now the link is UP and ready for data transmission.
	 * Here we perform Full-duplex data transmission:
	 * each side of the link sends one data frame and waits for the
	 * frame sent by the remote.
	 * The M_DATA message is used during data transfer.
	 */
	send_data(llc2_fd);
	recv_data(llc2_fd);

	/* Now the program sleeps for a while and then exits.
	 * It is possible to close the connection by using DL_DISCON
	 * primitives, and to unbind the stream by using DL_UNBIND primitives.
	 * This is done automatically if we just exit.
	 */
	sleep(10);
	exit(0);
}

void usage(void)
{
	printf("USAGE : llc2 <link_id> <device> 1 to initiate connection\n");
	printf("        llc2 <link_id> <device> 0 to receive connection\n");
	printf("\nEx : to build a LLC2 connection over ethernet between 2 stations use :\n");
	printf("     llc2 28 /dev/le 0 (on the first station)\n");
	printf(" and llc2 29 /dev/le 1 (on the other station)\n");
	exit(1);
}

/* These functions show how to build the stream architecture 
 * so that the LLC2 DLPI interface can be used on top of the 
 * LAN driver.
 * 
 * To use the LLC2 DLPI interface, the stream
 * architecture must be :
 *
 *           |
 *  ------------------
 *  | LLC2 driver     |
 *  ------------------
 *           |
 *  ------------------
 *  |   LAN  driver  |
 *  ------------------
 *
 */
int build_stream_arch(char * device, int link_id)
{
	int dev_fd, muxid;
	int llc2_fd;
	struct ll_snioc snioc;

	/* Open the LAN device. */
	dev_fd = open(device, O_RDWR);
	if (dev_fd < 0) {
                perror("Cannot open", device);
                exit(1);
        }

	/* Open the llc2 device to access the LLC2 DLPI interface. */
	llc2_fd = open("/dev/llc2", O_RDWR);
	if (llc2_fd < 0) {
                perror("Cannot open /dev/llc2");
                exit(1);
        }

	/* The LAN driver also implements the DLPI interface. 
	 * Attach the LAN driver to ppa 0 to use first instance of 
	 * the LAN driver (example /dev/le0).
	 * See the man pages for le(7) for more information.
	 */
        send_dl_attach(dev_fd, 0);

	/* I_LINK device file (and wanmod) underneath /dev/llc2. */
	if ((muxid = ioctl(llc2_fd, I_LINK, dev_fd)) < 0) {
		perror("	**ioctl(I_LINK)");
		exit (1);
	}

	/*
	 * The stream architecture is now constructed and can be used.
	 * The first step is to send the L_SETPPA ioctls in order to
	 * set the PPA/snid of the device stream under LLC2 to 'link_id'.
	 * This PPA can then be used in the DL_ATTACH ioctl to identify
	 * the link.
	 *
	 * Both L_SETPPA and DL_ATTACH ioctls are described
	 * in the X.25 Programming Guide.
	 */
	snioc.lli_type = LI_SPPA;
	snioc.lli_ppa = link_id;
	snioc.lli_index = muxid;
	snioc.lli_class = LC_LLC2;

	strio.ic_cmd = L_SETPPA;
	strio.ic_timout = -1;			/* Infinite timeout */
	strio.ic_len = sizeof(snioc);
	strio.ic_dp = (char *) &snioc;
	if (ioctl(llc2_fd, I_STR, (char *) &strio) < 0) {
		perror("	**ioctl(L_SETPPA)");
		exit (1);
	}

	/* The L_SETTUNE ioctl can now be used to configure LLC2 parameters.
	 * In that case we don't send it and therefore the default values
	 * will be used for this LLC2 link.
	 */

	return llc2_fd;
}

/* This function attaches a previously built stream architecture
 * using the ppa : link_id 
 */
void send_dl_attach(int fd, int link_id)
{
	dl_attach_req_t dlatt;
	int flags;

	dlatt.dl_primitive = DL_ATTACH_REQ;
	dlatt.dl_ppa = link_id;
	ctlbuf.buf = (char *)&dlatt;
        ctlbuf.len = sizeof(dlatt);
        if (putmsg(fd, &ctlbuf, (struct strbuf *)0, 0) < 0)
        {
               perror("DL_ATTACH put message failed");
               exit(1);
        }

	/* Wait for the OK_ACK message that should acknowledge the ATTACH request.
	 *
	 * The generic union structure DL_primitives is used to retrieve
	 * messages from the LAPB driver because several types of messages
	 * can be returned depending on the success or failure of the 
	 * operation.
	 */
	wait_dlpi_prim(fd, DL_OK_ACK, "DL_ATTACH_REQ");
}

void send_dl_bind(int fd, int initiator_flag)
{
	dl_bind_req_t dlbind;

	dlbind.dl_primitive = DL_BIND_REQ;
	dlbind.dl_sap = 0x34;        /* use LSAP 0x34 in this example */
	/* conind must be set to 0 to initiate connection
	 * and to the max number of connection to respond to connections
	 * in our case we just want to receive one connection so set it to 1
	 */
	dlbind.dl_max_conind = (initiator_flag ? 0 : 1);
	dlbind.dl_service_mode = DL_CODLS;
	dlbind.dl_conn_mgmt = 0;
	dlbind.dl_xidtest_flg = DL_AUTO_XID|DL_AUTO_TEST;
	ctlbuf.buf = (char *)&dlbind;
        ctlbuf.len = sizeof(dlbind);
        if (putmsg(fd, &ctlbuf, (struct strbuf *)0, 0) < 0)
        {
               perror("DL_BIND put message failed");
               exit (1);
        }

	/* Now wait for the DL_BIND_ACK primitive that should acknowledge
	 * the DL_BIND_REQ.
	 */
	wait_dlpi_prim(fd, DL_BIND_ACK, "DL_BIND_REQ");
}

void send_dl_connect_req(int fd)
{
	char conn_struct[100];
	dl_connect_req_t * dlconnreq_p;
	struct llc_dladdr * llc_dladdr_p;
	union DL_primitives dlprim;
	int flags=0;

	dlconnreq_p = (dl_connect_req_t *) conn_struct;
	llc_dladdr_p = (struct llc_dladdr *) (&(conn_struct[0]) + sizeof(dl_connect_req_t));
	dlconnreq_p->dl_primitive = DL_CONNECT_REQ;
	dlconnreq_p->dl_dest_addr_length = sizeof(struct llc_dladdr);
	
	/* This is the offset to the llc_dladdr structure from the 
	 * beginning of conn_struct.
	 */
	dlconnreq_p->dl_dest_addr_offset = sizeof(dl_connect_req_t);
	
	/* Now fill the llc_dladdr structure with the SAP and MAC addresses.
	 */
	llc_dladdr_p->dl_sap = 0x34;
	llc_dladdr_p->dl_mac[0] = 0x08;
	llc_dladdr_p->dl_mac[1] = 0x00;
	llc_dladdr_p->dl_mac[2] = 0x20;
	llc_dladdr_p->dl_mac[3] = 0x0e;
	llc_dladdr_p->dl_mac[4] = 0x4d;
	llc_dladdr_p->dl_mac[5] = 0xc3;

        ctlbuf.buf = (char *)&(conn_struct[0]);
        ctlbuf.len = sizeof(dl_connect_req_t)+sizeof(struct llc_dladdr);
        if (putmsg(fd, &ctlbuf, (struct strbuf *)0, 0) != 0)
        {
                perror("DL_CONNECT_REQ put message failed");
                exit(1);
        }
	
        ctlbuf.buf = (char *)&dlprim;
        ctlbuf.maxlen = sizeof(dlprim);
        flags = 0;
        if (getmsg(fd, &ctlbuf, (struct strbuf *)0, &flags)<0)
        {
                perror("DL_CONNECT get message failed");
                exit(1);
        }
	 
        if (dlprim.dl_primitive == DL_CONNECT_CON)
                printf("DL_CONNECT_CON received OK\n");
	else
	{
		printf("DL_CONNECT failed\n");
		if (dlprim.dl_primitive == DL_ERROR_ACK)
		{
			printf("DL_ERROR_ACK was received\n");
			printf("     with errno : %d\n",
						 dlprim.error_ack.dl_errno);
			}
		if (dlprim.dl_primitive == DL_DISCONNECT_IND)
		{
			printf("DL_DISCONNECT_IND was received\n");
			printf("     with reason : %d\n",
					dlprim.disconnect_ind.dl_reason);
		}
		exit (1);
	}
}

/* This function waits for a DL_CONNECT_IND (connection indication)
 * from the LLC2 layer, then responds with DL_CONNECT_RES to accept the
 * connection.
 */
void wait_and_ack_dl_connect_ind(int fd)
{
	union DL_primitives dlprim;
	dl_connect_res_t dlconnres;
	int flags=0;

	/* Now the program waits for the DL_CONNECT_IND primitive that should indicate 
	 * that the LLC2 has completed the connection phase.
	 */
        ctlbuf.buf = (char *)&dlprim;
        ctlbuf.maxlen = sizeof(dlprim);
 
        flags = 0;
 
        if (getmsg(fd, &ctlbuf, (struct strbuf *)0, &flags) != 0)
        {
                perror("DL_CONNECT_IND get message failed");
                exit(1);
        }
 
        if (dlprim.dl_primitive == DL_CONNECT_IND)
                printf("DL_CONNECT_IND received : OK\n");
	else
	{
		printf("DL_CONNECT failed\n");
		exit (1);
	}
	/* Respond to a CONNECT indication with a DL_CONNECT_RES message. */
	dlconnres.dl_primitive = DL_CONNECT_RES;
	dlconnres.dl_correlation = dlprim.connect_ind.dl_correlation;
	dlconnres.dl_resp_token = 0;
	ctlbuf.buf = (char *)&dlconnres;
        ctlbuf.len = sizeof(dlconnres);
        if (putmsg(fd, &ctlbuf, (struct strbuf *)0, 0) < 0)
        {
               perror("DL_CONNECT_RES put message failed");
               exit (1);
        }

	/* Wait for the OK_ACK message that should acknowledge
	 * the CONNECT_RES message.
	 */
	wait_dlpi_prim(fd, DL_OK_ACK, "DL_CONNECT_RES");
}

/* This function transfers a 100 octet long buffer to the remote
 * by sending a simple M_DATA message to the LAPB driver.
 * No ack message is used in this phase.
 */
void send_data(int fd)
{
	struct strbuf databuf;
	char buf[100];

	databuf.len = 100;
	databuf.buf = buf;

	/* send a M_DATA Message */
	if (putmsg(fd, (struct strbuf *)0, &databuf, 0) < 0)
	{
		perror("M_DATA put message failed");
		exit(1);
	}

	/* release allocated buffer */
	printf("M_DATA (100 bytes) sent : OK\n");
}

/* This function waits for data.
 * The data are received within M_DATA messages.
 */
void recv_data(int fd)
{
	struct strbuf databuf;
	int flags = 0;
	/* It is the responsability of the application
	 * to decide and know the maximum length of data that can
	 * be received. Practically, this length can not be greater than
	 * the max length of the LAPB frames.
	 * In this example we know that the other side will never 
	 * send frames bigger than 100.
	 */
#define MAXDATALEN 100
	char buf[MAXDATALEN];

	databuf.buf = buf;
	databuf.maxlen = MAXDATALEN;
	databuf.len = 0;

	/* wait for a M_DATA Message */
	if (getmsg(fd, 0, &databuf, &flags) < 0)
	{
		perror("M_DATA receiving failed");
		exit(1);
	}

	/* In this example no checking is done on the data received. */

	printf("M_DATA (%d bytes) received : OK\n", databuf.len);
}

/* This function will wait for a DLPI message (exp_prim) acknowledging
 * a previously sent DLPI primitive (ack_prim_str).
 */
void wait_dlpi_prim(int fd, int exp_prim, char *ack_prim_str)
{
	union DL_primitives dlprim;
	struct strbuf ctlbuf;
	int flags=0;

	/* wait for exp_prim */
	ctlbuf.buf = (char *)&dlprim;
	ctlbuf.maxlen = sizeof(dlprim);
	if (getmsg(fd, &ctlbuf, (struct strbuf *)0, &flags) != 0)
	{
		printf("%s : ", ack_prim_str);
		perror(" get message failed");
		exit(1);
	}
 
	if(dlprim.dl_primitive == exp_prim)
	{
		/* success */
		printf("%s : OK \n",ack_prim_str);
		return;
	}

	switch(dlprim.dl_primitive)
	{
	    case DL_ERROR_ACK:
		printf("Error after %s : DL_ERROR_ACK was received\n",
					ack_prim_str);
		printf("     dl_errno is : %d\n", dlprim.error_ack.dl_errno);
		exit(1);
	    default:
		printf("Error after %s : Unexpected primitive received: %d\n",
					 dlprim.dl_primitive);
		exit(1);
	}
}

